# Perl 5 extension to SICStus Prolog v2.1pl9
# Version 1.0
# Author: Charles Bailey  bailey@genetics.upenn.edu

=head1 TITLE

SICStus - a Perl interface to SICStus Prolog runtime systems

=head1 SYNOPSIS

require SICStus;
$sp = new SICStus;
$sp->rtinit();
$boolean = $sp_qload($file);
$boolean = $sp->eval($string);
$boolean = $sp->assert($string);
$boolean = $sp->query($predicate,$arg,$arg,...);

$query = new SICStus::Query $predicate,$arg,$arg,...;
$boolean = $query->next_solution();
$success = $query->cut_query();
$query->close_query();

=head1 DESCRIPTION

The SICStus extension provides access from Perl to a SICStus Prolog
engine running in the same process.  Arbitrary strings can be passed
to Prolog for evaluation as if they were typed into the interpreter,
compiled Prolog object code can be loaded, and known predicates can
be satisfied directly either once or repeatedly.  The Prolog engine
is a runtime system, with all the rights, privileges, and immunities -
and liabilities - thereunto appertaining.

=head2 Starting Prolog

The SICStus extension is designed so that most functions can be
invoked by method calls using a C<SICStus> object.  Therefore, you
can use C<require> or C<use> to pull in F<SICStus.pm>; nothing is
imported into the caller's namespace.

Before you begin using Prolog, you need to call the C<rtinit> function
to initialize the Prolog engine and load the necessary glue code.  At
the moment, it doesn't matter whether you use a method call like

    require SICStus;
    $sp = new SICStus;
    $sp->rtinit();

or a simple function call like

    require SICStus;
    &SICStus::rtinit;

since there are no arguments required.  In the future, however, this
may change; in particular, if it becomes possible to create multiple
Prolog interpreters, it will be necessary to pass the interpreter
object to C<rtinit>.

=head2 Loading Prolog files

Since the Prolog engine is a runtime system, only Prolog object files
(F<.ql> files) can be loaded directly.  This is accomplished via the
C<qload> method call, which takes the path to the file as its
argument.

=head2 Evaling strings in Prolog

Arbitrary strings may be passed to Prolog using the C<eval> method
call. The strings are evaluated as though they had been typed into the
Prolog interpreter.  It is important that the string you pass to
Prolog represent a complete clause (i.e. end in a '.'), or the
Prolog engine will hang, waiting for the clause to be completed.

If the clause is satisfied, C<eval> returns 1.  If satisfaction of the
goal caused Prolog to instantiate any unbound terms, Perl scalars of
the same name are set to the values of the instantiated terms.  (This
happens in the package from which C<eval> was called.)  For instance,
if you said C<$sp->eval('home(north_pole,Who).');>, the Perl scalar
C<$Who> would be set to "Santa Claus" when the query succeeded.

If a query fails, C<eval> returns 0, and no Perl scalars are set.  If
Prolog returns an exception, Perl converts it into a Perl exception
whose value is the name of the Prolog exception term (as if Prolog had
called C<die "$exception_name"> in Perl).

The C<assert> and C<retract> methods are provided as shortcuts for the
Prolog predicates of the same name.  C<$sp->assert($fact)> and
C<$sp->retract($fiction)> are identical to
C<$sp->eval('assert('.$fact.').')> and
C<$sp->eval('retract('.$fiction.').')>, respectively.

=head2 Single satisfaction of a named predicate

You can ask to satisfy a predicate by name via the C<query> method,
passing a set of Perl scalars as arguments to the query.  The arity of
the predicate is inferred from the number of arguments passed.  (The
first parameter to the C<query> call is the name of the predicate; the
rest of the parameters are interpreted as arguments to the predicate.)

If a given argument is a defined Perl scalar, its value is passed to
Prolog as an instantiated term.  By default, Perl scalars which were
last used as numbers are passed to Prolog as floating point values,
and Perl scalars which were last used as strings are passed to Prolog
as character lists.  If you want to 'cast' a value to a given Prolog
type, you can use the following classes to generate objects of a
specific type, and pass the object to C<query> in place of the
original scalar:

=over 4

=item $obj = new SICStus::Atom $var
The string value of $var will be passed to Prolog as an atom.

=item $obj = new SICStus::Float $var
The numeric value of $var will be passed to Prolog as a floating point
number.

=item $obj = new SICStus::Integer $var
The numeric value of $var will be passed to Prolog as an integer.

=item $obj = new SICStus::List $var
The string value of $var will be passed to Prolog as a character list.

=back

If a given argument is has the value C<undef>, it is passed to Prolog
as an uninstantiated variable.  If it is instantiated during the
successful satisfaction of the query, the Perl scalar is set to the
value of the Prolog term.  Prolog integers and floating point numbers
are passed to Perl as numeric values, while atoms and character lists
are passed to Perl as strings.  If the Prolog term is a compound term
other than a character list, and the principal functor isn't ".", the
the corresponding Perl scalar is set to the name of the functor.  If a
functor other than "." is encountered when converting a compound term
which began as a character list (i.e. its principal functor was "."),
the corresponding Perl scalar is set to C<undef>.  I realize that this
is broken; it may get better in the future.

C<query> returns 1 if the goal was satisfied, and 0 if it failed.  If
Prolog returns an exception, Perl converts it into a Perl exception
whose value is the name of the Prolog exception term (as if Prolog had
called C<die "$exception_name"> in Perl).

=head2 Multiple satisfactions of a named predicate

You can repeatedly try to resatisfy a given predicate using the
methods in the C<SICStus::Query> package.  The query must first be set
up by a call to the C<new> method, which returns a query object for
use in later calls.  (You must also have initialized the Prolog engine
via a call to C<SICStus::rtinit>.)  The calling sequence is identical
to that for the C<query> method: C<$obj = new SICStus::Query
$pred,$arg,$arg,...>, and the same rules are followed for passing
arguments to Prolog.  The Perl scalars passed as arguments are bound
into the query.  C<new> returns 0 if there are no parameters, and
raises an exception if Prolog cannot set up the query.

Once the query is opened, you can attempt to satisfy it using the
C<next_solution> method, which takes no arguments.  If it succeeds, it
returns 1, and the Perl scalars bound into the query as arguments are
set to the values of the corresponding Prolog terms, according to the
rules specified above for the C<query> method in the C<SICStus>
package.  Each time a new solution is found, the Perl scalars are
reset; in between queries, you can do anything you like with them.  If
the Perl variables go out of scope, you lose access to the
corresponding Prolog terms, but no other harm is done.  If there are
no more solutions, C<next_solution> returns 0, and automatically
closes the query.  If Prolog returns an exception, Perl automatically
closes eth query, and then converts the Prolog exception into a Perl
exception whose value is the name of the Prolog exception term (as if
Prolog had called C<die "$exception_name"> in Perl).  If you finish
with a query before you run out of solutions, you have to close it
yourself, either by calling the C<close_query> method, or by
C<undef>ing the query object.

You may have multiple SICStus::Query queries open at any time, but
they must be strictly nested, that is, you must close a given query
before you close any queries which were opened before that query.

=head1 FILES

=item F<$PERL5LIB/SICStus.pm>  - Perl module for this extension
=item F<$PERL5LIB/SICSperl.ql> - Prolog glue for communication with Perl

=head1 BUGS

Compound terms could be expanded to Prolog 'canonical' representation.

It might be nice to add the full stop to eval'd strings so SICStus
doesn't hang if the user forgets it.

If a user closes a SICStus::Query out of order, we could mark it and
save it until the time is right.

=head1 AUTHOR

Charles Bailey  <bailey@genetics.upenn.edu>

=head1 COPYRIGHT

This work is copyright 1995 by Charles Bailey.  All rights reserved.
Permission is granted to use and copy this work under the same term as
Perl itself (See the file Copying in the Perl distribution for
details.)

NO WARRANTY is made that this software will function in any particular
manner or be fit for any purpose.  The author, the University of
Pennsylvania, and its affiliates will not be liable for any damages
whatsoever, including but not limited to direct, indirect, and
consequential damages, arising from the use of this software.

=cut

package SICStus;

use Carp;
require DynaLoader;
@ISA = qw( DynaLoader );

# At the moment, just a stub to return a handle for use in method
# calls.  It may do something real in the future if it becomes
# possible to start multiple Prolog engines.

sub new { my($self); bless \$self; }

sub rtinit {
  &SICStus::initialize or croak("Can't initialize SICStus runtime engine");
  &develinit;
}

sub develinit {
  my($qlfile);

  foreach $dir (@INC) {
    if (-f "$dir/SICSperl.ql") {
      $file = "$dir/SICSperl.ql";
      last;
    }
  }
  &SICStus::load($file) or croak("Can't load SICStus glue routines");

}

sub assert {
  my($self, $clause) = @_;
  $self->eval('assert('.$clause.').');
}

sub retract {
  my($self, $clause) = @_;
  $self->eval('retract('.$clause.').');
}

sub qload {
  my($self,$file);
  &SICStus::load($file);
}

bootstrap SICStus;

# Dummy packages to 'cast' Perl scalars to a particular Prolog type.
# The C code recognizes the blessed references and converts the
# values appropriately.

package SICStus::Atom;
sub new { my($self,$val) = @_; bless \($val || ''); }

package SICStus::Float;
sub new { my($self,$val) = @_; bless \($val || 0.0); }

package SICStus::Integer;
sub new { my($self,$val) = @_; bless \($val || 0); }

package SICStus::List;
sub new { my($self,$val) = @_; bless \($val || ''); }


package SICStus::Query;

use Carp;

sub new {
  my($ref) = 0;

  if (@_) { open_query(@_); }
  else { bless \$ref; }

}

# Fallback method to allow user to set up query after creating
# an empty SICStus::Query object.  Deprecated already.
sub set_query {
  my($self) = shift;

  if ($self) { croak "Can't override active query"; }
  $self = open_query(@_);

}
  
sub DESTROY {
  my($self) = shift;
  close_query($self);
}

1;
